@@ -79,6 +79,7 @@ class RunnerManager(kind: String, val fileManager: FileManager, params: TestRunP
 
   val compileMgr = new CompileManager(fileManager)
   fileManager.CLASSPATH += File.pathSeparator + PathSettings.scalaCheck
+  fileManager.CLASSPATH += File.pathSeparator + PathSettings.diffUtils // needed to put diffutils on test/partest's classpath
 
   private def compareFiles(f1: File, f2: File): String =
     try fileManager.compareFiles(f1, f2)
@@ -817,7 +818,7 @@ class RunnerManager(kind: String, val fileManager: FileManager, params: TestRunP
         if (passed exists (x => !x)) {
           if (fileManager.showDiff || isPartestDebug)
             NestUI.normal(testDiff)
-          else if (fileManager.showLog)
+          if (fileManager.showLog)
             showLog(logFile)
         }
       }
